package com.sillycycle.bagleyd.triangles;

/*
 * @(#)TrianglesCanvas.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>TrianglesCanvas</code> class does all the graphics for
 * <code>TrianglesFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class TrianglesCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 256;
	static final int MAX_SLICES = 10;

	static final int TR = 0;
	static final int RIGHT = 1;
	static final int BR = 2;
	static final int BL = 3;
	static final int LEFT = 4;
	static final int TL = 5;
	static final int TOP = 6;
	static final int BOTTOM = 7;
	static final int COORD = 6;
	static final int TRBL = 0;
	static final int TLBR = 1;
	static final int ROW = 2;
	static final int ROW_TYPES = 3;
	static final int DOWN = 0;
	static final int UP = 1;
	static final int MAX_ORIENT = 2;
	static final int BLOCKED = -3;
	static final int SPACE = -2;
	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "triangles.dat";
	static final String SYMBOL = ":";

	Color       background;
	Color       tileColor;
	int[]       tileOfPosition = null;
	int[]       startPosition = null;
	int[]       currentRow = new int[ROW_TYPES];
	int	 currentPosition = -1, currentPositionOrient;
	int	 currentSpace = -1;
	int	 currentDirection, lastDirection;
	int[]       spacePosition = new int[MAX_ORIENT];
	int[]       startSpace = new int[MAX_ORIENT];
	int[][]     spaceRow = new int[MAX_ORIENT][ROW_TYPES];
	int[][]     startRow = new int[MAX_ORIENT][ROW_TYPES];
	boolean     started, cheat, corners;
	int	 base;
	int	 delay/*, numSlices*/;
	int	 sizeX, sizeY, sizeSize, prevSizeSize;
	int	 pixX, pixY, pixNo;
	Point       coreSize;
	Point       offset;
	Point       tileSize, puzzleSize;
	Point       delta, puzzleOffset;

	JFrame      frame = null;
	int	 randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     movePaint = false, scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allTilesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false;
	boolean     focus = true;
	Random      generator;
	transient TrianglesStack undo = null, redo = null;
	String      stringSave = null;
	String[]    token = null;

	Point[][] triangleUnit = new Point[MAX_ORIENT][ROW_TYPES + 1];
	Point[] hexagonUnit = new Point[7];
	int[][] triangleListX = new int[MAX_ORIENT][ROW_TYPES + 1];
	int[][] triangleListY = new int[MAX_ORIENT][ROW_TYPES + 1];
	int[] hexagonListX = new int[7];
	int[] hexagonListY = new int[7];

	@Override
	protected String paramString() {
		return super.paramString() +
		",corners=" + corners +
		",sizeX=" + sizeX +
		",sizeY=" + sizeY +
		",base=" + base +
		",tileColor=" + tileColor +
		",background=" + background +
		",delay=" + delay;
	}

	void checkTiles() {
		if (sizeX < TrianglesInterface.MIN_TILES) {
			System.out.println(sizeX +
			" number of triangles in a row incorrect, must be at least " +
			TrianglesInterface.MIN_TILES +
			", defaulting to " +
			TrianglesInterface.DEFAULT_TILESX);
			sizeX = TrianglesInterface.DEFAULT_TILESX;
		}
		if (sizeY < TrianglesInterface.MIN_TILES) {
			System.out.println(sizeY +
			" number of triangles in a row incorrect, must be at least " +
			TrianglesInterface.MIN_TILES +
			", defaulting to " +
			TrianglesInterface.DEFAULT_TILESY);
			sizeY = TrianglesInterface.DEFAULT_TILESY;
		}
		if (base < TrianglesInterface.MIN_BASE ||
				base > TrianglesInterface.MAX_BASE) {
			System.out.println(base +
			" base incorrect, must be between " +
			TrianglesInterface.MIN_BASE +
			" and " +
			TrianglesInterface.MAX_BASE +
			" inclusive, defaulting to " +
			TrianglesInterface.DEFAULT_BASE);
			base = TrianglesInterface.DEFAULT_BASE;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	/* This is fast for small i, a -1 is returned for negative i. */
	static int sqrt(int i) {
		int j = 0;

		while (j * j <= i)
			j++;
		return (j - 1);
	}

	int fromRow(int row) {
		return (sizeX + row) * (sizeX + row)
			- (sizeX - 1) * (sizeX - 1);
	}

	int positionFromRow(int rowPosition, int posRow) {
		return rowPosition + fromRow(posRow);
	}

	int positionInRow(int row, int trbl, int tlbr) {
		return sizeX + trbl - tlbr + row - 1;
	}

	int toPosition(int row, int trbl, int tlbr) {
		return positionFromRow(positionInRow(row, trbl, tlbr), row - 1);
	}

	int toOrient(int row, int trbl, int tlbr) {
		return ((positionInRow(row, trbl, tlbr) % 2 == 1) ? DOWN : UP);
	}

	static int oppositeGrav(int grav) {
		return (grav == UP) ? DOWN : UP;
	}

	int toRow(int pos) {
		return sqrt(pos + (sizeX - 1) * (sizeX - 1)) - sizeX + 1;
	}

	/* Passing row so there is no sqrt calculation again */
	int toTrBl(int pos, int posRow) {
		return (pos - fromRow(posRow - 1)) >> 1;
	}

	int toTlBr(int pos, int posRow) {
		return (fromRow(posRow) - pos - 1) >> 1;
	}

	static int toDirHeading(int direction) {
		return (direction > 0 && direction < 4) ? UP : DOWN;
	}

	static int toRowType(int direction) {
		switch (direction) {
		case TR:
		case BL:
			return TRBL;
		case TL:
		case BR:
			return TLBR;
		case LEFT:
		case RIGHT:
			return ROW;
		default:
			return -1;
		}
	}

	int cartesianX(int pos, int row) {
		return (pos - fromRow(row - 1) - row + sizeY)
			* offset.x / 2 + delta.x / 2 + puzzleOffset.x;
	}

	int cartesianY(int row, int orient) {
		return ((orient == UP) ? 0 : tileSize.y) +
			row * offset.y + delta.y + 2 + puzzleOffset.y;
	}

	int tileNFromSpace(int n, int rowType, int orient, int direction) {
		int pos = spacePosition[orient];

		if (rowType == ROW) /* This one is easy */
			return (pos + ((direction == UP) ? -n : n));
		int row = toRow(pos);
		int trbl = toTrBl(pos, row);
		int tlbr = toTlBr(pos, row);
		int offset1, offset2;

		if (direction == UP) {
			offset1 = -((n >> 1) + (n & 1));
			offset2 = -(n >> 1);
		} else {
			offset1 = n >> 1;
			offset2 = (n >> 1) + (n & 1);
		}
		switch (rowType) {
		case TRBL:
			return ((orient == DOWN) ?
				toPosition(row + offset1,
					trbl, tlbr + offset2) :
				toPosition(row + offset2,
					trbl, tlbr + offset1));
		case TLBR:
			return ((orient == DOWN) ?
				toPosition(row + offset1,
					trbl + offset2, tlbr) :
				toPosition(row + offset2,
					trbl + offset1, tlbr));
		default: /* ROW */
			return (pos + ((direction == UP) ? -n : n));
		}
	}

	int tile1From(int pos, int rowType) {
		int row = toRow(pos);
		int trbl = toTrBl(pos, row);
		int tlbr = toTlBr(pos, row);
		int orient = toOrient(row, trbl, tlbr);

		if (orient == UP)
			switch(rowType) {
			case TR:
				if (row != sizeY - 1)
					return toPosition(row + 1, trbl, tlbr);
				else
					return -1;
			case RIGHT:
				if (trbl != 0)
					return pos - 1;
				else
				return -1;
			case BR:
				if (trbl != 0)
					return pos - 1;
				else
					return -1;
			case BL:
				if (tlbr != 0)
					return pos + 1;
				else
					return -1;
			case LEFT:
				if (tlbr != 0)
					return pos + 1;
				else
					return -1;
			case TL:
				if (row != sizeY - 1)
					return toPosition(row + 1, trbl, tlbr);
				else
					return -1;
			default:
				return -1;
			}
		else
			switch(rowType) {
			case TR:
			case RIGHT:
				return pos - 1;
			case BR:
			case BL:
				return toPosition(row - 1, trbl, tlbr);
			case LEFT:
			case TL:
				return pos + 1;
			default:
				return -1;
			}
	}

	int nextNoCornDir(int pos) {
		int row = toRow(pos);
		int trbl, tlbr;

		if (spaceRow[UP][ROW] == row) {
			if (spacePosition[UP] == pos - 1)
				return LEFT;
			else if (spacePosition[UP] == pos + 1)
				return RIGHT;
			else
				return -1;
		}
		trbl = toTrBl(pos, row);
		tlbr = toTlBr(pos, row);
		if (spaceRow[UP][TRBL] == trbl &&
				spaceRow[UP][TLBR] == tlbr) {
			if (spaceRow[UP][ROW] == row - 1)
				return TL; /* or TR */
			else if (spaceRow[UP][ROW] == row + 1)
				return BL; /* or BR */
			else
				return -1;
		}
		return -1;
	}

	int tileNextToSpace(int rowType, int orient, int direction) {
		if (direction == UP) {
			switch (rowType) {
			case TRBL:
				return ((orient == UP) ? spacePosition[orient] + 1 :
					spacePosition[orient] -
					2 * spaceRow[orient][ROW]);
			case TLBR:
				return ((orient == UP) ? spacePosition[orient] - 1 :
					spacePosition[orient] -
					2 * spaceRow[orient][ROW]);
			default: /* ROW */
				return (spacePosition[orient] - 1);
			}
		}
		/* direction == DOWN */
		switch (rowType) {
		case TRBL:
			return ((orient == DOWN) ? spacePosition[orient] - 1 :
				spacePosition[orient] +
				2 * (spaceRow[orient][ROW] + 1));
		case TLBR:
			return ((orient == DOWN) ? spacePosition[orient] + 1 :
				spacePosition[orient] +
				2 * (spaceRow[orient][ROW] + 1));
		default: /* ROW */
			return (spacePosition[orient] + 1);
		}
	}

	public boolean checkSolved() {
		int i;

		for (i = 1; i < sizeSize - ((corners) ?
				TrianglesInterface.CORNERS :
				TrianglesInterface.NOCORN); i++)
			if (tileOfPosition[i - 1] != i)
				return false;
		started = false;
		return true;
	}

	boolean bresenhamLine(int x0, int y0, int x1, int y1) {
		int pix = 0;
		int ex = x1 - x0;
		int ey = y1 - y0;
		int dx, dy, error;

		pixNo++;
		if (ex > 0) {
			dx = 1;
		} else if (ex < 0) {
			dx = -1;
			ex = -ex;
		} else {
			dx = 0;
		}
		if (ey > 0) {
			dy = 1;
		} else if (ey < 0) {
			dy = -1;
			ey = -ey;
		} else {
			dy = 0;
		}
		pixX = x0;
		pixY = y0;
		if (ex > ey) {
			error = 2 * ey - ex;
			while (pix != pixNo) {
				if (error >= 0) {
					error -= 2 * ex;
					pixY += dy;
				}
				error += 2 * ey;
				pixX += dx;
				pix++;
				if (pixX == x1)
					return true;
			}
			return false;
		}
		error = 2 * ex - ey;
		while (pix != pixNo) {
			if (error >= 0) {
				error -= 2 * ey;
				pixX += dx;
			}
			error += 2 * ex;
			pixY += dy;
			pix++;
			if (pixY == y1)
				return true;
		}
		return false;
	}


	static void fill3DTriangle(Graphics g, int startX, int startY,
			int[] listX, int[] listY, boolean raised,
			boolean orient) {
		int n = 3;
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		Color c = g.getColor();
		Color brighter = c.brighter();
		Color darker = c.darker();

		if (!raised) {
			g.setColor(darker);
		}
		g.fillPolygon(tempListX, tempListY, n);
		g.setColor((raised ^ orient) ? ((raised) ? brighter : c) :
			darker);
		g.drawLine(tempListX[0], tempListY[0],
			tempListX[1], tempListY[1]);
		g.drawLine(tempListX[1], tempListY[1],
			tempListX[2], tempListY[2]);
		if (orient) {
			g.drawLine(tempListX[0], tempListY[0] + 1,
				tempListX[1] - 1, tempListY[1]);
			g.drawLine(tempListX[1] - 1, tempListY[1] - 1,
				tempListX[2] + 1, tempListY[2] - 1);
		} else {
			g.drawLine(tempListX[0], tempListY[0] - 1,
				tempListX[1] + 1, tempListY[1]);
			g.drawLine(tempListX[1] + 1, tempListY[1] + 1,
				tempListX[2] - 1, tempListY[2] + 1);
		}
		g.setColor(raised ^ orient ? darker : ((raised) ? brighter :
			c));
		g.drawLine(tempListX[2], tempListY[2],
			tempListX[0], tempListY[0]);
		if (orient) {
			g.drawLine(tempListX[2] + 1, tempListY[2],
				tempListX[0], tempListY[0] + 1);
		} else {
			g.drawLine(tempListX[2] - 1, tempListY[2],
				tempListX[0], tempListY[0] - 1);
		}
		g.setColor(c);
	}

	static void fill3DHexagon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, boolean raised) {
		int n = 6;
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		Color c = g.getColor();
		Color brighter = c.brighter();
		Color darker = c.darker();

		if (!raised) {
			g.setColor(darker);
		}
		g.fillPolygon(tempListX, tempListY, n);
		g.setColor((raised) ? c : darker);
		g.drawLine(tempListX[4], tempListY[4],
			tempListX[5], tempListY[5]);
		g.drawLine(tempListX[1], tempListY[1],
			tempListX[2], tempListY[2]);
		g.setColor((raised) ? brighter : darker);
		g.drawLine(tempListX[5], tempListY[5],
			tempListX[0], tempListY[0]);
		g.drawLine(tempListX[0] - 1, tempListY[0],
			tempListX[1] - 1, tempListY[1]);
		g.drawLine(tempListX[0] - 1, tempListY[0] + 1,
			tempListX[1] - 1, tempListY[1]);
		/*g.drawLine(tempListX[5] + 1, tempListY[5],
			tempListX[0] + 1, tempListY[0]);
		g.drawLine(tempListX[5] + 2, tempListY[5],
			tempListX[0] + 2, tempListY[0]);*/
		g.setColor((raised) ? darker : c);
		g.drawLine(tempListX[2], tempListY[2],
			tempListX[3], tempListY[3]);
		g.drawLine(tempListX[3], tempListY[3],
			tempListX[4], tempListY[4]);
		g.drawLine(tempListX[3], tempListY[3] - 1,
			tempListX[4], tempListY[4] - 1);
		/*g.drawLine(tempListX[2] - 1, tempListY[2],
			tempListX[3] - 1, tempListY[3]);
		g.drawLine(tempListX[3] - 2, tempListY[2],
			tempListX[3] - 2, tempListY[3]);*/
	}

	/* void erase3DPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n,
			boolean orient) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.setColor(background);
		g.fillPolygon(tempListX, tempListY, n);
		g.drawLine(tempListX[0], tempListY[0],
			tempListX[1], tempListY[1]);
		g.drawLine(tempListX[1], tempListY[1],
			tempListX[2], tempListY[2]);
		if (orient) {
			g.drawLine(tempListX[0], tempListY[0] + 1,
				tempListX[1] - 1, tempListY[1]);
			g.drawLine(tempListX[1] - 1, tempListY[1] - 1,
				tempListX[2] + 1, tempListY[2] - 1);
		} else {
			g.drawLine(tempListX[0], tempListY[0] - 1,
				tempListX[1] + 1, tempListY[1]);
			g.drawLine(tempListX[1] + 1, tempListY[1] + 1,
				tempListX[2] - 1, tempListY[2] + 1);
		}
		g.drawLine(tempListX[2], tempListY[2],
			tempListX[0], tempListY[0]);
		if (orient) {
			g.drawLine(tempListX[2] + 1, tempListY[2],
				tempListX[0], tempListY[0] + 1);
		} else {
			g.drawLine(tempListX[2] - 1, tempListY[2],
				tempListX[0], tempListY[0] - 1);
		}
	} */

	static void drawTriangleShadow(Graphics g, int startX, int startY,
			int[] listX, int[] listY, boolean orient) {
		int n = 3;
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		if (orient) {
			g.drawLine(tempListX[2], tempListY[2],
				tempListX[0], tempListY[0]);
			g.drawLine(tempListX[2] + 1, tempListY[2],
				tempListX[0], tempListY[0] + 1);
		} else {
			g.drawLine(tempListX[0], tempListY[0],
				tempListX[1], tempListY[1]);
			g.drawLine(tempListX[1], tempListY[1],
				tempListX[2], tempListY[2]);
		}
	}

	static void drawHexagonShadow(Graphics g, int startX, int startY,
			int[] listX, int[] listY) {
		int n = 6;
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.drawLine(tempListX[4], tempListY[4],
			tempListX[5], tempListY[5]);
		g.drawLine(tempListX[5], tempListY[5],
			tempListX[0], tempListY[0]);
		g.drawLine(tempListX[0], tempListY[0],
			tempListX[1], tempListY[1]);
		g.drawLine(tempListX[5] + 1, tempListY[5],
			tempListX[0] + 1, tempListY[0]);
		g.drawLine(tempListX[1], tempListY[1],
			tempListX[1], tempListY[1] + 2);
	}

	static void fillPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.fillPolygon(tempListX, tempListY, n);
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawTile(int pos, int orient, boolean blank, boolean erase,
			int pressedOffset, int offsetX, int offsetY) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			int dx, dy, row = toRow(pos);

			dx = cartesianX(pos, row) + offsetX + pressedOffset;
			dy = cartesianY(row, orient) + offsetY + pressedOffset;

			Color fillColor, textColor;

			if (erase) {
				fillColor = background;
				textColor = background;
			} else {
				fillColor = tileColor;
				textColor = background.darker();
			}
			if (!corners) {
				dy += offset.x / 4 + 2 + ((orient == UP) ? 0 :
					-4 * tileSize.y / 3 - 1) +
					(tileSize.y + delta.y) / 3;
			}
			if (blank) {
				g.setColor(fillColor);
				if (corners)
					fillPolygon(g,
						dx, dy,
						triangleListX[orient],
						triangleListY[orient], 3);
				else
					fillPolygon(g,
						dx, dy,
						hexagonListX,
						hexagonListY, 6);
			} else {
				if (pressedOffset != 0) {
					g.setColor(fillColor.darker());
					if (corners)
						drawTriangleShadow(g,
							dx - pressedOffset,
							dy - pressedOffset,
							triangleListX[orient],
							triangleListY[orient],
							orient == UP);
					else
						drawHexagonShadow(g,
							dx - pressedOffset,
							dy - pressedOffset,
							hexagonListX,
							hexagonListY);
				}
				g.setColor(fillColor);
				if (corners)
					fill3DTriangle(g,
						dx, dy,
						triangleListX[orient],
						triangleListY[orient],
						pressedOffset == 0, orient == UP);
				else
					fill3DHexagon(g,
						dx, dy,
						hexagonListX,
						hexagonListY,
						pressedOffset == 0);
			}
			if (!blank) {
				int i, tile = tileOfPosition[pos];
				int digitOffsetX = 0, digitOffsetY, fontSize;

				g.setColor(textColor);
				fontSize = java.lang.Math.min(tileSize.x,
					tileSize.y) / 2;
				if (!corners)
					fontSize = fontSize * 3 / 4;
				g.setFont(new Font("arial", Font.PLAIN,
					fontSize));
				i = tile;
				while (i >= 1) {
					i /= base;
					digitOffsetX += (fontSize / 3);
				}
				digitOffsetY = (orient == UP) ?
					fontSize / 2 + 2 + tileSize.y / 10:
					-fontSize / 2 - 4 * tileSize.y / 5;
				if (!corners)
					digitOffsetY += -delta.y / 2 +
						((orient == UP) ? -5 * tileSize.y / 8 :
						5 * tileSize.y / 8);
				g.drawString(Integer.toString(tile, base).toUpperCase(),
					dx - digitOffsetX + tileSize.x / 25,
					dy + tileSize.y / 2 + digitOffsetY);
			}
		} finally {
			g.dispose();
		}
	}

	void drawAllTiles() {
		for (int pos = 0; pos < sizeSize; pos++) {
			int row = toRow(pos);
			int trbl = toTrBl(pos, row);
			int tlbr = toTlBr(pos, row);
			int orient = toOrient(row, trbl, tlbr);

			drawTile(pos, orient, (tileOfPosition[pos] <= 0),
				(tileOfPosition[pos] <= 0), 0, 0, 0);
		}
	}

	int movableCornerTile() {
		int rowType = BLOCKED, l;
		/* Are the spaces in a "row" with the mouse click?
		   (If two, then one clicked on a space). */

		for (l = 0; l < ROW_TYPES; l++) {
			if (currentRow[l] == spaceRow[DOWN][l] &&
					currentRow[l] == spaceRow[UP][l]) {
				if (rowType == BLOCKED) {
					rowType = l;
				} else {
					return SPACE;
				}
			}
		}
		return rowType;
	}

	int movableNoCornTile() {
		int rowType;

		if (currentPosition == spacePosition[UP])
			return SPACE;
		rowType = nextNoCornDir(currentPosition);
		if (rowType < 0)
			return BLOCKED;
		return rowType;
	}

	int countTiles(int dir, int rowType) {
		switch (rowType) {
		case TLBR:
			return spaceRow[dir][TRBL] - currentRow[TRBL] +
				spaceRow[oppositeGrav(dir)][ROW] -
				currentRow[ROW];
		case TRBL:
			return spaceRow[dir][TLBR] - currentRow[TLBR] +
				spaceRow[oppositeGrav(dir)][ROW] -
				currentRow[ROW];
		case ROW:
			return spaceRow[dir][TRBL] - currentRow[TRBL] +
				currentRow[TLBR] -
				spaceRow[oppositeGrav(dir)][TLBR];
		default:
			System.out.println("countTiles: rowType " + rowType);
		}
		return 0;
	}

	static int dirTiles(int dir, int rowType) {
		switch (rowType) {
		case TLBR:
			return (dir == UP) ? TL : BR;
		case TRBL:
			return (dir == UP) ? TR : BL;
		case ROW:
			return (dir == UP) ? LEFT : RIGHT;
		default:
			System.out.println("dirTiles: rowType " + rowType);
		}
		return dir;
	}

	void animateCornerSlide(int dir, int fast, boolean logMoves) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			int aTile, numTiles, direction = dir, rowType;
			int fillBeginPos, fillBeginPos2, fillBeginRow;
			int fillBeginOrient, fillBeginOrientOp;
			int pos, pos2, posNext, orient, orientNext;

			rowType = movableCornerTile();
			if (rowType < 0) {
				System.out.println(
					"animateCornerSlide: rowType " + rowType);
				numTiles = 0;
			} else {
				numTiles = countTiles(dir, rowType);
				direction = dirTiles(dir, rowType);
			}
			if (numTiles < 0)
				numTiles = -numTiles;
			if (dir == UP)
				orient = (spacePosition[UP] +
					((rowType == TRBL) ? 2 : 0) >
					spacePosition[DOWN]) ? UP : DOWN;
			else
				orient = (spacePosition[UP] +
					((rowType == TRBL) ? 1 : 0) <
					spacePosition[DOWN]) ? UP : DOWN;
			fillBeginPos = tileNFromSpace(numTiles, rowType,
				orient, oppositeGrav(dir));
			fillBeginRow = toRow(fillBeginPos);
			fillBeginOrient = toOrient(fillBeginRow,
				toTrBl(fillBeginPos, fillBeginRow),
				toTlBr(fillBeginPos, fillBeginRow));
			fillBeginOrientOp = (fillBeginOrient == UP) ?
				DOWN : UP;
			fillBeginPos2 = spacePosition[fillBeginOrientOp];
			{
				int fromdx, fromdy;
				int todx, tody;
				int dx, dy, row;
				int firstRow, firstOrient;
				int[] rowPos = new int[numTiles + 2];
				int i, k;
				boolean slideDone = false;

				pixNo = 0;
				pixX = 0;
				pixY = 0;
				rowPos[2] = tileNFromSpace(1, rowType, orient,
					oppositeGrav(dir));
				firstRow = toRow(rowPos[2]);
				firstOrient = toOrient(firstRow,
					toTrBl(rowPos[2], firstRow),
					toTlBr(rowPos[2], firstRow));
				rowPos[0] = spacePosition[firstOrient];
				rowPos[1] = spacePosition[(firstOrient == UP) ?
					DOWN : UP];
				for (i = 1; i < numTiles; i++) {
					rowPos[i + 2] = tileNFromSpace(i + 1,
						rowType, orient,
						oppositeGrav(dir));
				}
				row = toRow(rowPos[2]);
				fromdx = cartesianX(rowPos[2], row);
				fromdy = cartesianY(row, firstOrient);
				row = toRow(rowPos[0]);
				todx = cartesianX(rowPos[0], row);
				tody = cartesianY(row, firstOrient);
				dx = todx - fromdx;
				dy = tody - fromdy;
				i = 0;
				while (!slideDone) {
					for (k = 2; k < numTiles + 2; k++)
						drawTile(rowPos[k],
							(firstOrient + k) & 1,
							true, true, 0,
							pixX, pixY);
					slideDone = bresenhamLine(0, 0, dx, dy);
					for (k = 2; k < numTiles + 2; k++)
						drawTile(rowPos[k],
							(firstOrient + k) & 1,
							false, false, 0,
							pixX, pixY);
					if (i % 8 == 0) {
						try {
							Thread.sleep(delay / fast);
						} catch (InterruptedException e) {
							// ignore
						}
					}
					i++;
				}
			}
			for (aTile = 0; aTile < numTiles; aTile++) {
				if (logMoves) {
					((TrianglesFrame) frame).callback(
						TrianglesInterface.ACTION_MOVED);
					undo.setMove(direction);
					redo.flushMoves();
				}
				posNext = tileNFromSpace(aTile + 1, rowType,
					orient, oppositeGrav(dir));
				orientNext = ((aTile & 1) == 1) ? orient :
					oppositeGrav(orient);
				if (orientNext == fillBeginOrientOp)
					fillBeginPos2 = posNext;
			}
			pos2 = spacePosition[orient];
			pos = spacePosition[oppositeGrav(orient)];
			for (aTile = 0; aTile < numTiles; aTile++) {
				posNext = tileNFromSpace(aTile + 1, rowType,
					orient, oppositeGrav(dir));
				tileOfPosition[pos] = tileOfPosition[posNext];
				pos = pos2;
				pos2 = posNext;
			}
			tileOfPosition[fillBeginPos] =
				spacePosition[fillBeginOrient];
			tileOfPosition[fillBeginPos2] =
				spacePosition[fillBeginOrientOp];
			spacePosition[fillBeginOrient] = fillBeginPos;
			spacePosition[fillBeginOrientOp] = fillBeginPos2;
			if (fillBeginOrient == UP) {
				tileOfPosition[fillBeginPos] = 0;
				tileOfPosition[fillBeginPos2] = -1;
			} else {
				tileOfPosition[fillBeginPos] = -1;
				tileOfPosition[fillBeginPos2] = 0;
			}
			spaceRow[fillBeginOrient][ROW] = toRow(fillBeginPos);
			spaceRow[fillBeginOrient][TRBL] = toTrBl(fillBeginPos,
				spaceRow[fillBeginOrient][ROW]);
			spaceRow[fillBeginOrient][TLBR] = toTlBr(fillBeginPos,
				spaceRow[fillBeginOrient][ROW]);
			spaceRow[fillBeginOrientOp][ROW] = toRow(fillBeginPos2);
			spaceRow[fillBeginOrientOp][TRBL] = toTrBl(fillBeginPos2,
				spaceRow[fillBeginOrientOp][ROW]);
			spaceRow[fillBeginOrientOp][TLBR] = toTlBr(fillBeginPos2,
				spaceRow[fillBeginOrientOp][ROW]);
		} finally {
			g.dispose();
		}
	}

	void animateNoCornSlide(int fast, boolean logMoves) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			int direction;
			int fillBeginPos, fillBeginRow;
			int pos, posNext;

			direction = movableNoCornTile();
			if (direction < 0) {
				System.out.println(
					"animateNoCornSlide: direction " + direction);
				return;
			}
			fillBeginPos = tile1From(spacePosition[UP], direction);
			fillBeginRow = toRow(fillBeginPos);
			{
				int fromdx, fromdy;
				int todx, tody;
				int dx, dy;
				int firstOrient;
				int i;
				boolean slideDone = false;

				pixNo = 0;
				pixX = 0;
				pixY = 0;
				firstOrient = toOrient(fillBeginRow,
					toTrBl(fillBeginPos, fillBeginRow),
					toTlBr(fillBeginPos, fillBeginRow));
				fromdx = cartesianX(fillBeginPos, fillBeginRow);
				fromdy = cartesianY(fillBeginRow, firstOrient);
				todx = cartesianX(spacePosition[UP],
					spaceRow[UP][ROW]);
				tody = cartesianY(spaceRow[UP][ROW],
					((firstOrient == 0) ? 1 : 0));
				if (firstOrient == 0)
					tody += (4 * tileSize.y / 3 + 2);
				else
					fromdy += (4 * tileSize.y / 3 + 2);
				dx = todx - fromdx;
				dy = tody - fromdy;
				i = 0;
				while (!slideDone) {
					drawTile(fillBeginPos,
						firstOrient,
						true, true, 0,
						pixX, pixY);
					drawTile(fillBeginPos,
						firstOrient,
						true, true, 0,
						pixX + 1, pixY + 1);
					slideDone = bresenhamLine(0, 0, dx, dy);
					drawTile(fillBeginPos,
						firstOrient,
						false, false, 0,
						pixX, pixY);
					if (i % 8 == 0) {
						try {
							Thread.sleep(delay / fast);
						} catch (InterruptedException e) {
							// ignore
						}
					}
					i++;
				}
			}
			if (logMoves) {
				((TrianglesFrame) frame).callback(
					TrianglesInterface.ACTION_MOVED);
				undo.setMove(direction);
				redo.flushMoves();
			}
			posNext = tile1From(spacePosition[UP], direction);
			pos = spacePosition[UP];
			tileOfPosition[pos] = tileOfPosition[posNext];
			pos = posNext;
			tileOfPosition[fillBeginPos] =
				spacePosition[UP];
			spacePosition[UP] = fillBeginPos;
			tileOfPosition[fillBeginPos] = 0;
			spaceRow[UP][ROW] = toRow(fillBeginPos);
			spaceRow[UP][TRBL] = toTrBl(fillBeginPos,
				spaceRow[UP][ROW]);
			spaceRow[UP][TLBR] = toTlBr(fillBeginPos,
				spaceRow[UP][ROW]);
		} finally {
			g.dispose();
		}
	}

	void resetTiles() {
		int pos;

		sizeSize = sizeY * sizeY + 2 * sizeY * (sizeX - 1);
		prevSizeSize = (sizeY - 1) * (sizeY - 1)
			+ 2 * (sizeY - 1) * (sizeX - 1);
		tileOfPosition = new int[sizeSize];
		startPosition = new int[sizeSize];
		/*for (i = 0; i < sizeSize - 2; i++)
			tileOfPosition[i] = i + 1;
		spacePosition[UP] = sizeSize - 1;
		spaceRow[UP][TRBL] = size - 1; // i
		spaceRow[UP][TLBR] = 0; // j
		spaceRow[UP][ROW] = size - 1; // k
		tileOfPosition[sizeSize - 1] = 0;
		if (size > 1) {
			spacePosition[DOWN] = sizeSize - 2;
			spaceRow[DOWN][TRBL] = size - 2; // i
			spaceRow[DOWN][TLBR] = 0; // j
			spaceRow[DOWN][ROW] = size - 1; // k
			tileOfPosition[sizeSize - 2] = -1;
		}*/
		if (sizeY < 2) {
			for (pos = 0; pos < sizeSize; pos++) {
				if (pos == 0)
					tileOfPosition[pos] = 0;
				else if (corners && pos == 1)
					tileOfPosition[pos] = -1;
			else
				tileOfPosition[pos] = pos - ((corners) ? 2 : 1) + 1;
			}
			spacePosition[UP] = 0;
			spaceRow[UP][TRBL] = 0; /* i */
			spaceRow[UP][TLBR] = sizeX - 1; /* j */
			spaceRow[UP][ROW] = 0; /* k */
			if (corners) {
				if (sizeX > 1) {
					spacePosition[DOWN] = 1;
					spaceRow[DOWN][TLBR] = sizeX - 2; /* j */
				} else {
					spacePosition[DOWN] = 0;
					spaceRow[DOWN][TLBR] = 0; /* j */
				}
				spaceRow[DOWN][TRBL] = 0; /* i */
				spaceRow[DOWN][ROW] = 0; /* k */
			}
		} else {
			if (!corners && sizeX == 1 && sizeY == 3) {
				tileOfPosition[0] = 1;
				tileOfPosition[1] = 4;
				tileOfPosition[2] = 7;
				tileOfPosition[3] = 6;
				tileOfPosition[4] = 0;
				tileOfPosition[5] = 5;
				tileOfPosition[6] = 3;
				tileOfPosition[7] = 2;
				tileOfPosition[8] = 8;
			} else if (!corners && sizeX == 2 && sizeY == 2) {
				tileOfPosition[0] = 3;
				tileOfPosition[1] = 6;
				tileOfPosition[2] = 5;
				tileOfPosition[3] = 0;
				tileOfPosition[4] = 4;
				tileOfPosition[5] = 2;
				tileOfPosition[6] = 1;
				tileOfPosition[7] = 7;
			} else {
				int parityCalc = sizeY % 4;
				boolean paritySwap = (parityCalc == 0) ||
					(parityCalc == 1 && sizeX % 2 == 1) ||
					(parityCalc == 3 && sizeX % 2 == 0);
				for (pos = 0; pos < sizeSize; pos++) {
					int row, trbl, tlbr;

					row = toRow(pos);
					trbl = toTrBl(pos, row);
					tlbr = toTlBr(pos, row);
					if (pos == prevSizeSize)
						tileOfPosition[pos] = 0;
					else if (pos == prevSizeSize + 1) {
						if (corners)
							tileOfPosition[pos] = -1;
						else
							tileOfPosition[pos] = pos;
					} else if (!corners && pos == sizeSize - 1)
						tileOfPosition[pos] = pos;
					else if (!corners && paritySwap && pos == sizeSize - 2)
						tileOfPosition[pos] = pos;
					else if (!corners && paritySwap && pos == prevSizeSize + 2)
						tileOfPosition[pos] = pos;
					else
						tileOfPosition[pos] = toPosition(row, tlbr, trbl - 1);
				}
			}
			spacePosition[UP] = fromRow(sizeY - 2);
			spaceRow[UP][TRBL] = 0; /* i */
			spaceRow[UP][TLBR] = sizeX + sizeY - 2; /* j */
			spaceRow[UP][ROW] = sizeY - 1; /* k */
			if (corners) {
				spacePosition[DOWN] = spacePosition[UP] + 1;
				spaceRow[DOWN][TRBL] = spaceRow[UP][TRBL]; /* i */
				spaceRow[DOWN][TLBR] = spaceRow[UP][TLBR] - 1; /* j */
				spaceRow[DOWN][ROW] = spaceRow[UP][ROW]; /* k */			}
		}
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		randomPaints = 0;
	}

	public void drawFrame(Graphics g, boolean show) {
		int sumX, sumY, sumX1, sumX2, offsetX, offsetY;
		int[] tempListX = new int[4];
		int[] tempListY = new int[4];

		if (g == null)
			return;
		offsetX = puzzleOffset.x;
		offsetY = puzzleOffset.y;
		sumX = (sizeX + sizeY - 1) * offset.x + delta.x + 1;
		sumY = sizeY * offset.y + delta.y + 1;
		sumX1 = (sizeY * offset.x + delta.x) / 2 + offsetX;
		sumX2 = sumX1 + (sizeX - 1) * offset.x + 1;
		sumX += offsetX;
		sumY += offsetY;
		g.setColor(background.darker().darker());
		if (offsetY > 0)
			g.fillRect(0, 0, coreSize.x, offsetY);
		if (sumY < coreSize.y)
			g.fillRect(0, sumY + 1, coreSize.x,
				coreSize.y - sumY);
		if (offsetX > 0)
			g.fillRect(0, 0, offsetX, coreSize.y);
		if (sumX - 2 < coreSize.x)
			g.fillRect(sumX - 2, 0, coreSize.x - sumX + 2,
				coreSize.y);
		tempListX[0] = offsetX - 1;
		tempListY[0] = offsetY;
		tempListX[1] = sumX1 - 3;
		tempListY[1] = offsetY;
		tempListX[2] = offsetX - 2;
		tempListY[2] = sumY;
		g.fillPolygon(tempListX, tempListY, 3);
		tempListX[0] = sumX - 2;
		tempListY[0] = offsetY - 1;
		tempListX[1] = sumX - 2;
		tempListY[1] = sumY + 1;
		tempListX[2] = sumX2;
		tempListY[2] = offsetY - 1;
		g.fillPolygon(tempListX, tempListY, 3);
		if (show) { /* 3DFrame */
			Color brighterColor = (focus) ?
				background.brighter() : background;
			Color darkerColor = (focus) ?
				background.darker() : background;

			tempListX[0] = offsetX;
			tempListY[0] = sumY + 1;
			tempListX[1] = sumX1 - 3;
			tempListY[1] = offsetY - 1;
			g.setColor(darkerColor);
			g.drawPolyline(tempListX, tempListY, 2);
			tempListX[0] = sumX2 - 1;
			tempListY[0] = offsetY - 1;
			tempListX[1] = sumX - 1;
			tempListY[1] = sumY + 1;
			tempListX[2] = offsetX + 1;
			tempListY[2] = sumY + 1;
			g.setColor(brighterColor);
			g.drawPolyline(tempListX, tempListY, 3);
			tempListX[0] = offsetX + 1;
			tempListY[0] = sumY + 1;
			tempListX[1] = sumX1 - 2;
			tempListY[1] = offsetY - 1;
			g.setColor(darkerColor);
			g.drawPolyline(tempListX, tempListY, 2);
			tempListX[0] = sumX2 - 1;
			tempListY[0] = offsetY;
			tempListX[1] = sumX - 1;
			tempListY[1] = sumY;
			tempListX[2] = offsetX + 1;
			tempListY[2] = sumY;
			g.setColor(brighterColor);
			g.drawPolyline(tempListX, tempListY, 3);
			tempListX[0] = offsetX + 1;
			tempListY[0] = sumY;
			tempListX[1] = sumX1 - 2;
			tempListY[1] = offsetY;
			g.setColor(darkerColor);
			g.drawPolyline(tempListX, tempListY, 2);
			tempListX[0] = offsetX;
			tempListY[0] = sumY;
			tempListX[1] = sumX1 - 3;
			tempListY[1] = offsetY;
			g.setColor(darkerColor);
			g.drawPolyline(tempListX, tempListY, 2);
			tempListX[0] = sumX2;
			tempListY[0] = offsetY - 1;
			tempListX[1] = sumX - 2;
			tempListY[1] = sumY + 1;
			tempListX[2] = offsetX;
			tempListY[2] = sumY + 1;
			g.setColor(brighterColor);
			g.drawPolyline(tempListX, tempListY, 3);
		} else {
			tempListX[0] = sumX1 - 1;
			tempListY[0] = offsetY;
			tempListX[1] = offsetX + 1;
			tempListY[1] = sumY;
			tempListX[2] = sumX - 1;
			tempListY[2] = sumY;
			tempListX[3] = sumX2 - 2;
			tempListY[3] = offsetY;
			g.setColor(background);
			g.fillPolygon(tempListX, tempListY, 4);
		}
	}

	public void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void moveNoTiles() {
		((TrianglesFrame) frame).callback(
			TrianglesInterface.ACTION_IGNORE);
	}

	void moveTiles(int from, int orient, int fast) {
		if (corners)
			moveCornerTiles(from, orient, fast);
		else
			moveNoCornTiles(from, fast);
	}

	void moveCornerTiles(int from, int orient, int fast) {
		if (fast != INSTANT && delay > 0) {
			currentPosition = from;
			currentRow[ROW] = toRow(from);
			currentRow[TRBL] = toTrBl(from, currentRow[ROW]);
			currentRow[TLBR] = toTlBr(from, currentRow[ROW]);
			currentPositionOrient = toOrient(currentRow[ROW],
				currentRow[TRBL], currentRow[TLBR]);
			if (currentPosition <
					spacePosition[currentPositionOrient]) {
				animateCornerSlide(DOWN, fast, false);
			} else if (currentPosition >
					spacePosition[currentPositionOrient]) {
				animateCornerSlide(UP, fast, false);
			} else
				return;
			currentPosition = -1;
		} else {
			int tempTile = tileOfPosition[from];

			tileOfPosition[from] =
				tileOfPosition[spacePosition[orient]];
			tileOfPosition[spacePosition[orient]] = tempTile;
			drawTile(spacePosition[orient], orient, false, false,
				0, 0, 0);
			spacePosition[orient] = from;
			spaceRow[orient][ROW] = toRow(from);
			spaceRow[orient][TRBL] =
				toTrBl(from, spaceRow[orient][ROW]);
			spaceRow[orient][TLBR] =
				toTlBr(from, spaceRow[orient][ROW]);
			drawTile(spacePosition[orient], orient, true, true,
				0, 0, 0);
		}
		if (((TrianglesFrame) frame).getToggleSound())
			((TrianglesFrame) frame).playBumpAudio();
	}

	void moveNoCornTiles(int from, int fast) {
		int orient = UP, drawOrient;

		if (fast != INSTANT && delay > 0) {
			currentPosition = from;
			currentRow[ROW] = toRow(from);
			currentRow[TRBL] = toTrBl(from, currentRow[ROW]);
			currentRow[TLBR] = toTlBr(from, currentRow[ROW]);
			if (currentPosition < spacePosition[UP]) {
				animateNoCornSlide(fast, false);
			} else if (currentPosition > spacePosition[UP]) {
				animateNoCornSlide(fast, false);
			} else
				return;
			currentPosition = -1;
		} else {
			int tempTile = tileOfPosition[from];

			tileOfPosition[from] = tileOfPosition[spacePosition[orient]];
			tileOfPosition[spacePosition[orient]] = tempTile;
			drawOrient = toOrient(spaceRow[orient][ROW],
				spaceRow[orient][TRBL],
				spaceRow[orient][TLBR]);
			drawTile(spacePosition[orient], drawOrient,
				false, false, 0, 0, 0);
			spacePosition[orient] = from;
			spaceRow[orient][ROW] = toRow(from);
			spaceRow[orient][TRBL] =
				toTrBl(from, spaceRow[orient][ROW]);
			spaceRow[orient][TLBR] =
				toTlBr(from, spaceRow[orient][ROW]);
			drawOrient = toOrient(spaceRow[orient][ROW],
				spaceRow[orient][TRBL],
				spaceRow[orient][TLBR]);
			drawTile(spacePosition[orient], drawOrient,
				true, true, 0, 0, 0);
			drawTile(spacePosition[orient], drawOrient,
				true, true, 1, 0, 0);
		}
		if (((TrianglesFrame) frame).getToggleSound())
			((TrianglesFrame) frame).playBumpAudio();
	}

	int cornerDirToOrient(int direction) {
		switch (direction) {
		case TR:
			return (spacePosition[UP] + 2 >
				spacePosition[DOWN]) ? UP : DOWN;
		case RIGHT:
		case BR:
			return (spacePosition[UP] <
				spacePosition[DOWN]) ? UP : DOWN;
		case BL:
			return (spacePosition[UP] + 1 <
				spacePosition[DOWN]) ? UP : DOWN;
		case LEFT:
		case TL:
			return (spacePosition[UP] >
				spacePosition[DOWN]) ? UP : DOWN;
		default:
			return -1;
		}
	}

	boolean checkingMoveCornerTilesDir(int direction) {
		switch (direction) {
		case TR:
			return (spaceRow[UP][TRBL] == spaceRow[DOWN][TRBL] &&
					spaceRow[UP][ROW] != sizeY - 1);
		case RIGHT:
			return (spaceRow[UP][ROW] == spaceRow[DOWN][ROW] &&
					spaceRow[UP][TRBL] != 0);
		case BR:
			return (spaceRow[UP][TLBR] == spaceRow[DOWN][TLBR] &&
					spaceRow[UP][TRBL] != 0);
		case BL:
			return (spaceRow[UP][TRBL] == spaceRow[DOWN][TRBL] &&
					spaceRow[UP][TLBR] != 0);
		case LEFT:
			return (spaceRow[UP][ROW] == spaceRow[DOWN][ROW] &&
					spaceRow[UP][TLBR] != 0);
		case TL:
			return (spaceRow[UP][TLBR] == spaceRow[DOWN][TLBR] &&
					spaceRow[UP][ROW] != sizeY - 1);
		default:
			return false;
		}
	}

	boolean checkMoveCornerTilesDir(int direction) {
		if (checkingMoveCornerTilesDir(direction)) {
			currentPositionOrient = cornerDirToOrient(direction);
			int rowType = toRowType(direction);
			int heading = toDirHeading(direction);
			currentPosition = tileNextToSpace(rowType, currentPositionOrient, heading);
			currentPositionOrient = oppositeGrav(currentPositionOrient);
			return true;
		}
		return false;
	}

	boolean checkMoveNoCornTilesDir(int direction) {
		currentPosition = tile1From(spacePosition[UP], direction);
		if (currentPosition >= 0 && movableNoCornTile() >= 0)
			return true;
		return false;
	}

	boolean checkMoveTilesDir(int direction) {
		if (corners)
			return checkMoveCornerTilesDir(direction);
		return checkMoveNoCornTilesDir(direction);
	}

	boolean movePuzzleDir(int direction) {
		currentDirection = direction;
		if (direction == TOP || direction == BOTTOM) {
			if (corners)
				return false;
			if (spaceRow[UP][TRBL] + spaceRow[UP][TLBR] == spaceRow[UP][ROW])
				currentDirection = ((direction == TOP) ?
					TR /* or TL */ : BR /* or BL */);
			else
				return false;
		}
		if (checkMoveTilesDir(currentDirection)) {
			movePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	void movePuzzle(int direction, boolean control) {
		if (mouseDown || randomPaints != 0)
			return;
		if (stackPaint) {
			repaint();
			return;
		}
		if (control) {
			switch (direction) {
			case TL:
			case TOP:
			case TR:
				((TrianglesFrame) frame).callback(
					TrianglesInterface.ACTION_INCY);
				break;
			case RIGHT:
				((TrianglesFrame) frame).callback(
					TrianglesInterface.ACTION_INCX);
				break;
			case BL:
			case BOTTOM:
			case BR:
				if (sizeY <= TrianglesInterface.MIN_TILES)
					return;
				((TrianglesFrame) frame).callback(
					TrianglesInterface.ACTION_DECY);
				break;
			case LEFT:
				if (sizeX <= TrianglesInterface.MIN_TILES)
					return;
				((TrianglesFrame) frame).callback(
					TrianglesInterface.ACTION_DECX);
				break;
			default:
				System.out.println(
					"movePuzzle: should not get here");
			}
			return;
		}
		if (checkSolved()) {
			moveNoTiles();
			return;
		}
		if (!movePuzzleDir(direction)) {
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_BLOCKED);
		}
	}

	boolean exchangeTiles(int pos1, int pos2) {
		int tempTile;

		if ((tileOfPosition[pos1] <= 0) || (tileOfPosition[pos2] <= 0))
			return false;
		tempTile = tileOfPosition[pos1];
		tileOfPosition[pos1] = tileOfPosition[pos2];
		tileOfPosition[pos2] = tempTile;
		return true;
	}

	void discreteCornerMoves(int dir, int fast) {
		int rowType, orient, next;

		rowType = movableCornerTile();
		if (rowType < 0) {
			System.out.println("discreteCornerMoves: rowType " +
				rowType);
		}
		if (dir == DOWN) {
			orient = (spacePosition[UP] +
				((rowType == TRBL) ? 1 : 0) <
				spacePosition[DOWN]) ? UP : DOWN;
			next = tileNFromSpace(1, rowType, orient, UP);
			orient = oppositeGrav(orient);
			moveCornerTiles(next, orient, fast);
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_MOVED);
			switch (rowType) {
			case TLBR:
				undo.setMove(BR);
				redo.flushMoves();
				return;
			case TRBL:
				undo.setMove(BL);
				redo.flushMoves();
				return;
			case ROW:
				undo.setMove(RIGHT);
				redo.flushMoves();
				return;
			default:
				System.out.println(
					"discreteCornerMoves: rowType " +
					rowType);
				return;
			}
		}
		/* dir == UP */
		orient = (spacePosition[UP] +
			((rowType == TRBL) ? 2 : 0) >
			spacePosition[DOWN]) ? UP : DOWN;
		next = tileNFromSpace(1, rowType, orient, DOWN);
		orient = oppositeGrav(orient);
		moveCornerTiles(next, orient, fast);
		((TrianglesFrame) frame).callback(
			TrianglesInterface.ACTION_MOVED);
		switch (rowType) {
		case TLBR:
			undo.setMove(TL);
			redo.flushMoves();
			return;
		case TRBL:
			undo.setMove(TR);
			redo.flushMoves();
			return;
		case ROW:
			undo.setMove(LEFT);
			redo.flushMoves();
			return;
		default:
			System.out.println("discreteCornerMoves: rowType " +
				rowType);
		}
	}

	void discreteNoCornMoves(int dir, int fast) {
		int rowType, next;

		rowType = movableNoCornTile();
		if (rowType < 0) {
			System.out.println("discreteNoCornMoves: rowType " +
				rowType);
		}
		if (dir == DOWN) {
			next = tile1From(spacePosition[UP], rowType);
			moveNoCornTiles(next, fast);
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_MOVED);
			undo.setMove(rowType);
			redo.flushMoves();
			return;
		}
		/* dir == UP */
		next = tile1From(spacePosition[UP], rowType);
		moveNoCornTiles(next, fast);
		((TrianglesFrame) frame).callback(
			TrianglesInterface.ACTION_MOVED);
		undo.setMove(rowType);
		redo.flushMoves();
	}

	int positionToTile(int positionX, int positionY) {
		int sumY, sumX1, sumX2;
		int row, trbl, tlbr;
		int x = positionX, y = positionY;

		sumY = sizeY * offset.y + delta.y + 1;
		sumX1 = (sizeY * offset.x + delta.x) / 2;
		sumX2 = sumX1 + (sizeX - 1) * offset.x + 1;
		x -= puzzleOffset.x;
		y -= puzzleOffset.y;
		if (y < 0) /* above */
			return -2;
		if (y > sumY - delta.y) /* below */
			return -3;
		if (x * (sumY + delta.y) + y * (sumX1 + 1) <
				(sumX1 + 1) * (sumY + delta.y)) /* left */
			return -4;
		if (x * (sumY + delta.y) - y * (sumX2 - 1) >
				(sumX2 - 1) * (sumY + delta.y)) /* right */
			return -5;
		row = (y - delta.y) / offset.y;
		trbl = (x - sumX1 - 1 + row * offset.x / 2) / offset.x;
		trbl += ((x - (trbl + 1) * offset.x) *
			(sumY + delta.y) + y * (sumX1 + 1)) /
			((sumX1 + 1) * (sumY + delta.y));
		tlbr = (-x + sumX2 - 1 + row * offset.x / 2) /
			offset.x;
		tlbr += 1 + ((-x - (tlbr + 1) * offset.x) *
			(sumY + delta.y) + y * (sumX1 - 1)) /
			((sumX2 - 1) * (sumY + delta.y));
		if (row >= 0 && trbl >= 0 && tlbr >= 0 && row < sizeY &&
			trbl < sizeX + sizeY - 1 && tlbr < sizeX + sizeY - 1) {
			return toPosition(row, trbl, tlbr);
		}
		return -1;
	}

	void selectCornerTiles() {
		if (currentPosition < spacePosition[currentPositionOrient]) {
			if (delay > 0) {
				animateCornerSlide(DOWN, NORMAL, true);
				if (((TrianglesFrame) frame).getToggleSound())
					((TrianglesFrame) frame).playBumpAudio();
			} else {
				while (currentPosition <
						spacePosition[currentPositionOrient]) {
					discreteCornerMoves(DOWN, NORMAL);
				}
			}
		} else {
			if (delay > 0) {
				animateCornerSlide(UP, NORMAL, true);
				if (((TrianglesFrame) frame).getToggleSound())
					((TrianglesFrame) frame).playBumpAudio();
			} else {
				while (currentPosition >
						spacePosition[currentPositionOrient]) {
					discreteCornerMoves(UP, NORMAL);
				}
			}
		}
		if (checkSolved())
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_SOLVED);
	}

	void selectNoCornTiles() {
		if (currentPosition < 0 || currentSpace < 0)
			return;
		discreteNoCornMoves(((currentPosition < currentSpace) ? UP : DOWN),
			NORMAL);
		if (checkSolved())
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_SOLVED);
	}

	boolean isDeadCorner(int pos) {
		if (!corners) {
			if (tileOfPosition[pos] == 1 ||
					tileOfPosition[pos] == sizeSize - 1 ||
					tileOfPosition[pos] == prevSizeSize)
				return true;
		}
		return false;
	}

	void randomizeTiles() {
		// First interchange tiles but only with other tiles
		// of the same orientation and even number of times
		if (sizeY > 2) {
			int currentPos, randomPos, randomRow;
			int currentOrient = UP, randomOrient = DOWN;
			int step = 1, fin = 1, count = 0;

			for (currentPos = 0; currentPos < sizeSize;
					currentPos++) {
				if (isDeadCorner(currentPos)) {
					continue;
				}
				randomPos = currentPos;
				while (currentPos == randomPos ||
						currentOrient != randomOrient ||
						isDeadCorner(randomPos)) {
					randomPos = generator.nextInt(sizeSize);
					randomRow = toRow(randomPos);
					randomOrient = (((randomRow +
						toTrBl(randomPos, randomRow) +
						toTlBr(randomPos, randomRow))
						& 1) == 1) ? DOWN : UP;
				}
				if (exchangeTiles(currentPos, randomPos))
					count++;
				if (fin == currentPos + 1) {
					currentOrient = UP;
					step += 2;
					fin += step;
				} else
					currentOrient = (currentOrient == UP) ?
						DOWN : UP;
			}
			if ((count & 1) == 1 && !corners) {
				if (!exchangeTiles(1, 3) &&
						!exchangeTiles(5, 7))
				System.out.println("randomizeTiles: should not get here");
			}
			allTilesPaint = true;
			repaint();
		}
		// Now move the space around randomly
		if (sizeX + sizeY > 2) {
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_RESET);
			lastDirection = -1;
			randomPaints = java.lang.Math.min(sizeSize,
				MAX_ITERATIONS) + generator.nextInt(2);
			repaint();
		} else if (checkSolved())
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_SOLVED);
	}

	void randomizingTiles() {
		currentDirection = generator.nextInt(COORD);
		if ((currentDirection + COORD / 2) % COORD != lastDirection &&
				checkMoveTilesDir(currentDirection)) {
			moveTiles(currentPosition,
				currentPositionOrient, INSTANT);
			lastDirection = currentDirection;
			currentDirection = -1;
			randomPaints--;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_RANDOMIZE);
			started = false;
			currentPosition = -1;
			if (checkSolved())
				((TrianglesFrame) frame).callback(
					TrianglesInterface.ACTION_SOLVED);
		}
	}

	void resizeTiles() {
		int i, j;

		for (j = 0; j < MAX_ORIENT; j++)
			for (i = 0; i <= ROW_TYPES; i++) {
				triangleListX[j][i] = (tileSize.x / 2) *
					triangleUnit[j][i].x +
					((i > 0) ? triangleListX[j][i - 1] : 0);
				triangleListY[j][i] = tileSize.y *
					triangleUnit[j][i].y +
					((i > 0) ? triangleListY[j][i - 1] : 0);
			}
		for (i = 0; i <= 6; i++) {
			hexagonListX[i] = (tileSize.x / 6) *
				hexagonUnit[i].x +
				((i > 0) ? hexagonListX[i - 1] : 0);
			hexagonListY[i] = (tileSize.y / 3) *
				hexagonUnit[i].y +
				((i > 0) ? hexagonListY[i - 1] : 0);
		}

	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
		/* numSlices = ((delay < MAX_SLICES) ? delay + 1 :
			MAX_SLICES); */
	}

	public boolean getCheat() {
		return cheat;
	}

	public boolean setPuzzleSizeX(int newValue) {
		if (sizeX != newValue && newValue >=
				TrianglesInterface.MIN_TILES) {
			sizeX = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public boolean setPuzzleSizeY(int newValue) {
		if (sizeY != newValue && newValue >=
				TrianglesInterface.MIN_TILES) {
			sizeY = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleMode(boolean newValue) {
		if (corners != newValue) {
			corners = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((TrianglesFrame) frame).callback(
			TrianglesInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((TrianglesFrame) frame).callback(
			TrianglesInterface.ACTION_SPEED);
	}

	void showMessage(String msg) {
		((TrianglesFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		double sqrt_3 = 1.73205080756887729352744634150587237;
		Point tempSize;

		/* Determine size of client area */
		coreSize = new Point(getSize().width, getSize().height);
		delta = new Point(5, 3);
		offset = new Point(
			java.lang.Math.max((coreSize.x - delta.x) / (sizeX + sizeY - 1), 0),
			java.lang.Math.max((coreSize.y - 2 * delta.y) / sizeY, 0));
		tempSize = new Point(
			(int) (offset.y * 2.0 / sqrt_3),
			(int) (offset.x * sqrt_3 / 2.0));
		if (tempSize.y < offset.y)
			offset.y = tempSize.y;
		else /* tempSize.x <= offset.x */
			offset.x = tempSize.x;
		puzzleSize = new Point(offset.x * (sizeX + sizeY - 1) + delta.x + 2,
			offset.y * sizeY + delta.y + 2);
		puzzleOffset = new Point((coreSize.x - puzzleSize.x + 2) / 2,
			(coreSize.y - puzzleSize.y + 2) / 2);
		tileSize = new Point(java.lang.Math.max(offset.x - delta.x, 0),
			java.lang.Math.max(offset.y - delta.y, 0));
		resizeTiles();
	}

	public void initializePuzzle() {
		checkTiles();
		undo = new TrianglesStack();
		redo = new TrianglesStack();
		resetTiles();
		generator = new Random(System.nanoTime());
		triangleUnit[DOWN][0] = new Point(0, 0);
		triangleUnit[DOWN][1] = new Point(-1, -1);
		triangleUnit[DOWN][2] = new Point(2, 0);
		triangleUnit[DOWN][3] = new Point(-1, 1);
		triangleUnit[UP][0] = new Point(0, 0);
		triangleUnit[UP][1] = new Point(1, 1);
		triangleUnit[UP][2] = new Point(-2, 0);
		triangleUnit[UP][3] = new Point(1, -1);
		hexagonUnit[0] = new Point(1, -1);
		hexagonUnit[1] = new Point(1, 1);
		hexagonUnit[2] = new Point(-1, 1);
		hexagonUnit[3] = new Point(-2, 0);
		hexagonUnit[4] = new Point(-1, -1);
		hexagonUnit[5] = new Point(1, -1);
		hexagonUnit[6] = new Point(2, 0);
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		drawAllTiles();
	}

	void selectPuzzle(int x, int y) {
		currentPosition = positionToTile(x, y);
		if (currentPosition >= 0) {
			if (checkSolved()) {
				moveNoTiles();
				currentPosition = -1;
				return;
			}
			int row = toRow(currentPosition);
			currentRow[ROW] = row;
			currentRow[TRBL] = toTrBl(currentPosition, row);
			currentRow[TLBR] = toTlBr(currentPosition, row);
			currentPositionOrient = toOrient(row,
				currentRow[TRBL], currentRow[TLBR]);
			selectPaint = true;
			repaint();
		} else
			currentPosition = -1;
	}

	void releasePuzzle() {
		if (currentPosition == -1)
			return;
		currentSpace = 0;
		releasePaint = true;
		repaint();
	}

	public void clearPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		resetTiles();
		drawAllTiles();
		allTilesPaint = true;
		repaint();
		((TrianglesFrame) frame).callback(
			TrianglesInterface.ACTION_RESET);
	}


	public void clearWithQueryPuzzle() {
		if (!started)
			clearPuzzle();
	}

	public void randomizePuzzle() {
		if (!mouseDown)
			randomizeTiles();
	}

	void cornersPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((TrianglesFrame) frame).callback(
			TrianglesInterface.ACTION_CORNERS);
	}

	public void getPuzzle() {
		this.stringSave = ((TrianglesFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			int i, puzzleXDim, puzzleYDim, moves;
			boolean cornersValue;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			cornersValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.corners != cornersValue)
				((TrianglesFrame) frame).callback(
					TrianglesInterface.ACTION_CORNERS);
			lineData = token[++line];
			puzzleXDim = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (puzzleXDim >= TrianglesInterface.MIN_TILES) {
				for (i = this.sizeX; i < puzzleXDim; i++) {
					((TrianglesFrame) frame).callback(
						TrianglesInterface.ACTION_INCX);
				}
				for (i = this.sizeX; i > puzzleXDim; i--) {
					((TrianglesFrame) frame).callback(
						TrianglesInterface.ACTION_DECX);
				}
			} else {
				showMessage(source +
					" corrupted: sizeX " + puzzleXDim +
					" should be at least " +
					(TrianglesInterface.MIN_TILES +
					((cornersValue) ? 1 : 0)));
				return;
			}
			lineData = token[++line];
			puzzleYDim = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (puzzleYDim >= TrianglesInterface.MIN_TILES) {
				for (i = this.sizeY; i < puzzleYDim; i++) {
					((TrianglesFrame) frame).callback(
						TrianglesInterface.ACTION_INCY);
				}
				for (i = this.sizeY; i > puzzleYDim; i--) {
					((TrianglesFrame) frame).callback(
						TrianglesInterface.ACTION_DECY);
				}
			} else {
				showMessage(source +
					" corrupted: sizeY " + puzzleYDim +
					" should be at least " +
					TrianglesInterface.MIN_TILES);
				return;
			}
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			if (line < 0)
				return;
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_RESTORE);
			setStartPosition();
			allTilesPaint = true;
			repaint();
			if (!scanMoves(++line, moves))
				return;
			System.out.println(source + ": cornersValue " +
				((cornersValue) ? 1: 0) + ", sizeX " + puzzleXDim +
				", sizeY " + puzzleYDim + ", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found while reading");
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("corners").append(symbol).append(space)
			.append((corners) ? TrianglesInterface.CORNERS : TrianglesInterface.NOCORN).append(newline);
		sb.append("sizeX").append(symbol).append(space)
			.append(sizeX).append(newline);
		sb.append("sizeY").append(symbol).append(space)
			.append(sizeY).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((TrianglesFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	public void undoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!undo.madeMoves())
			return;
		undo.getMove();
		int direction = undo.readMoveDirection();

		redo.setMove(direction);
		// move in opposite direction
		direction = (direction + COORD / 2) % COORD;
		if (checkMoveTilesDir(direction)) {
			currentDirection = direction;
			stackPaint = true;
			repaint();
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_UNDO);
		}
	}

	public void redoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!redo.madeMoves())
			return;
		redo.getMove();
		int direction = redo.readMoveDirection();

		undo.setMove(direction);
		if (checkMoveTilesDir(direction)) {
			currentDirection = direction;
			stackPaint = true;
			repaint();
			((TrianglesFrame) frame).callback(
				TrianglesInterface.ACTION_REDO);
		}
	}

	public void solvePuzzle() {
		showMessage("Auto-solver: sorry, not implemented.");
	}

	void printPositions() {
		int r = 0;

		for (int i = 0; i < sizeSize; i++) {
			if (i == (r + sizeX - 1) * (r + sizeX - 1))
				for (int tab = 0; tab < sizeX + sizeY - r - 2; tab++)
					System.out.print("    ");
			int pos = startPosition[i];
			if (pos < 100)
				System.out.print(" ");
			if (pos < 10 && pos >= 0)
				System.out.print(" ");
			System.out.print(pos);
			if (i == (r + sizeX) * (r + sizeX) - 1) {
				r++;
				System.out.println();
			} else
				System.out.print(" ");
		}
		System.out.println();
	}

	void getStartPosition() {
		startSpace[UP] = spacePosition[UP];
		startSpace[DOWN] = spacePosition[DOWN];
		startRow[UP][TRBL] = spaceRow[UP][TRBL];
		startRow[UP][TLBR] = spaceRow[UP][TLBR];
		startRow[UP][ROW] = spaceRow[UP][ROW];
		startRow[DOWN][TRBL] = spaceRow[DOWN][TRBL];
		startRow[DOWN][TLBR] = spaceRow[DOWN][TLBR];
		startRow[DOWN][ROW] = spaceRow[DOWN][ROW];
		System.arraycopy(tileOfPosition, 0, startPosition, 0, sizeSize);
	}

	void setStartPosition() {
		spacePosition[UP] = startSpace[UP];
		spacePosition[DOWN] = startSpace[DOWN];
		spaceRow[UP][TRBL] = startRow[UP][TRBL];
		spaceRow[UP][TLBR] = startRow[UP][TLBR];
		spaceRow[UP][ROW] = startRow[UP][ROW];
		spaceRow[DOWN][TRBL] = startRow[DOWN][TRBL];
		spaceRow[DOWN][TLBR] = startRow[DOWN][TLBR];
		spaceRow[DOWN][ROW] = startRow[DOWN][ROW];
		System.arraycopy(startPosition, 0, tileOfPosition, 0, sizeSize);
		// drawAll
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int n = 0, r = 0;

			++line; //read startingPosition:
			lineData = token[++line];
			String[] st = lineData.split(space);
			int offset = ("".equals(st[0])) ? 1 : 0;
			for (int i = 0; i < sizeSize; i++) {
				int num = Integer.parseInt(st[n + offset].trim());
				startPosition[i] = num;
				if (num == 0)
					startSpace[UP] = i;
				else if (num == -1)
					startSpace[DOWN] = i;
				n++;
				if (i == (r + sizeX) * (r + sizeX) - 1 &&
						i != sizeSize - 1) {
					r++;
					n = 0;
					line++;
					if (line >= token.length) {
						System.out.println("Short data in StartPosition");
						return -1;
					}
					lineData = token[line];
					st = lineData.split(space);
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input in StartPosition");
			return -1;
		}
		for (int i = DOWN; i <= UP; i++) {
			startRow[i][ROW] = toRow(startSpace[i]);
			startRow[i][TRBL] = toTrBl(startSpace[i],
				startRow[i][ROW]);
			startRow[i][TLBR] = toTlBr(startSpace[i],
				startRow[i][ROW]);
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;

			++line; //read moves direction
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData = token[line];
				currentDirection = Integer.parseInt(
					lineData.substring(lineData.indexOf(symbol) +
					2));
				if (checkMoveTilesDir(currentDirection)) {
					scanPaint = true;
					paint(g);
				} else {
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
					currentDirection = -1;
					return false;
				}
			}
			if (l < moves) {
				showMessage("Corrupt input in scanMoves, only " + l + " moves");
			}
		} catch (NumberFormatException e) {
			e.printStackTrace();
			showMessage("Corrupt input found in scanMoves");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;
		int r = 0, oldR = -1;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int i = 0; i < sizeSize; i++) {
			if (oldR != r) {
				for (int tab = 0; tab < sizeY - r - 1; tab++)
					sb.append(space).append(space)
						.append(space).append(space);
				oldR = r;
			}
			int pos = startPosition[i];
			if (pos < 100)
				sb.append(space);
			if (pos < 10 && pos >= 0)
				sb.append(space);
			sb.append(space).append(pos);
			if (i == (r + 1) * (r + 1)
					+ 2 * (r + 1) * (sizeX - 1) - 1) {
				r++;
				sb.append(newline);
			}
		}
		sb.append(newline);
	}

	void printMoves(StringBuilder sb) {
		int direction;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space).append("direction").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			redo.setMove(direction);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(newline);
			undo.setMove(direction);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allTilesPaint &&
				!selectPaint && !releasePaint && !movePaint &&
				!scanPaint && !stackPaint && randomPaints == 0)
			resizePaint = true;
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allTilesPaint) {
			drawAllTiles();
			allTilesPaint = false;
		}
		if (selectPaint) {
			if (currentPosition != -1) {
				int rowType = (corners) ? movableCornerTile() :
				  movableNoCornTile();

				if (rowType < 0) {
				  drawTile(currentPosition,
				    currentPositionOrient,
				    rowType == SPACE,
				    false, 1, 0, 0);
				  try {
				    Thread.sleep((long) 3 * delay);
				  } catch (InterruptedException e) {
					  // ignore
				  }
				  drawTile(currentPosition,
				    currentPositionOrient,
				    true, true, 1, 0, 0);
				  if (rowType != SPACE)
				    drawTile(currentPosition,
				      currentPositionOrient,
				      false, false, 0, 0, 0);
				  ((TrianglesFrame) frame).callback(rowType);
				  currentPosition = -1;
				  mouseDown = false;
				} else
				  drawTile(currentPosition,
				    currentPositionOrient,
				    false, false, 1, 0, 0);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (currentPosition != -1) {
				drawTile(currentPosition,
					currentPositionOrient,
					true, true, 1, 0, 0);
				drawTile(currentPosition,
					currentPositionOrient,
					false, false, 0, 0, 0);
				if (currentSpace != -1)
					if (corners)
						selectCornerTiles();
					else
						selectNoCornTiles();
				currentPosition = -1;
			}
			releasePaint = false;
		}
		if (movePaint || scanPaint || stackPaint) {
			int fast;

			if (stackPaint)
				fast = DOUBLE;
			else if (movePaint)
				fast = NORMAL;
			else
				fast = INSTANT;
			moveTiles(currentPosition, currentPositionOrient, fast);
			started = true;
			if (!stackPaint && (movePaint || scanPaint)) {
				((TrianglesFrame) frame).callback(
					TrianglesInterface.ACTION_MOVED);
				undo.setMove(currentDirection);
				redo.flushMoves();
				if (checkSolved())
					((TrianglesFrame) frame).callback(
						TrianglesInterface.ACTION_SOLVED);
				movePaint = false;
				scanPaint = false;
			}
			stackPaint = false;
			currentDirection = -1;
			currentPosition = -1;
		}
		if (randomPaints > 0) {
			randomizingTiles();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}
